#!/usr/bin/perl
#  
# Copyright (c) 2003, 2004
#      Ben Summers.  All rights reserved.
#  
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All use of this software and associated advertising materials must 
#    display the following acknowledgement:
#        This product includes software developed by Ben Summers.
# 4. The names of the Authors may not be used to endorse or promote
#    products derived from this software without specific prior written
#    permission.
# 
# [Where legally impermissible the Authors do not disclaim liability for 
# direct physical injury or death caused solely by defects in the software 
# unless it is modified by a third party.]
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#  
#  
#  
use strict;

my @modules;
my %module_dependency;
my %module_library_link_opts;
my %header_dependency;


# note: Mac OS X resource forks and .DS_Store files are explicity ignored

print "Box build environment setup.\n\nChecking environment...\n";


my $implicit_dep = 'lib/common';

# work out platform variables
use BoxPlatform;

# don't allow old versions of openssl by default.
my $old_version_of_openssl_ok = 0;

# do command line arguments
my $compile_line_extra = '';
my $link_line_extra = '';
for(@ARGV)
{
	if($_ eq 'allow-old-openssl')
	{
		$old_version_of_openssl_ok = 1;
		next;
	}
	my ($k,$v) = split /:/,$_,2;
	if($k eq 'compile')
	{
		$compile_line_extra .= $v . ' ';
	}
	elsif($k eq 'link')
	{
		$link_line_extra .= $v . ' ';
	}
	elsif($k eq 'openssl')
	{
		# assume that the bin/lib/include dirs are under the specified path
		chop $v if ($v =~ /\/$/);
		$compile_line_extra = "-I$v/include $compile_line_extra"; 
		$link_line_extra    = "-L$v/lib $link_line_extra";
	}
	else
	{
		die "invalid option $_ specified on command line"
	}
}


# ---- LINUX start --------------------------------------------------------------------------
# do configuration on Linux... find out what's available, and configure appropraitely.
my $pseudo_db_lib = 'NONE';
my $pseudo_readline_lib = 'NONE';
if($build_os eq 'Linux')
{
	my ($linux_readline_h, $linux_db_h);
	my $db_ver = -1;
	
	open H_FILES,"find /usr/include |" or die "Can't open find";
	while(<H_FILES>)
	{
		chomp;
		
		if(m~/readline.h\Z~)
		{
			$linux_readline_h = strip_h_name($_);
			$pseudo_readline_lib = 'readline';
		}
		if(m~/db(|_1\d+).h\Z~)
		{
			# version?
			my $dv = 0;
			$dv = $1 if m~(\d+)/db~;
			my $fn = $_;

			# check this file mentions dbopen
			open DB_H,$_ or die "Can't open $_";
			my $found = 0;
			while(<DB_H>)
			{
				if(m/dbopen/)
				{
					$found = 1;
					last;
				}
			}
			close DB_H;
			next unless $found;

			# see if this version works
			print("Checking db version... (ignore error messages)\n");
			my $db_h = strip_h_name($fn);
			my $db_lib = 'db';
			$db_lib .= $db_ver if $db_ver > 0;
			if(!check_db_lib($db_h, $db_lib))
			{
				# try another
				my $db_lib = 'db';
				next unless check_db_lib($db_h, $db_lib)
			}

	 		# good version?
			if($dv > $db_ver)
			{
				$linux_db_h = $db_h;
				$db_ver = $dv;
				$pseudo_db_lib = $db_lib;
			}
		}
	}
	print "Finished checking versions\n";
	close H_FILES;

	# write the platform file accordingly and tell the user what's happened.
	open LIN,">lib/common/_linux_platform.h" or die "Can't open file for writing";
	if($linux_readline_h eq '')
	{
		print("---------------------\nWARNING: readline isn't installed\n---------------------\n");
		print LIN "#define PLATFORM_READLINE_NOT_SUPPORTED\n";

	}
	else
	{
		open RL_H,">lib/common/_linux_readline.h" or die "Can't open file for writing";
		print RL_H "#include <$linux_readline_h>\n";
		my $hist = $linux_readline_h;
		$hist =~ s/readline\.h/history.h/;
		print RL_H "#include <$hist>\n";
		close RL_H;
	}
	if($linux_db_h eq '')
	{
		print("---------------------\nWARNING: db is not installed -- will run in reduced efficiency mode without it.\n---------------------\n");
		print LIN "#define PLATFORM_BERKELEY_DB_NOT_SUPPORTED\n";
	}	
	else
	{
		open DB_H,">lib/common/_linux_db.h" or die "Can't open file for writing";
		print DB_H "#include <$linux_db_h>\n";
		close DB_H;
	}
	close LIN;
}
sub strip_h_name
{
	my $i = $_[0];
	$i =~ s~\A/usr/include/~~;
	return $i;
}
sub check_db_lib
{
	my ($h,$db) = @_;
	open TESTDBLIB,">testdblib.cpp" or die "Can't open openssl test file for writing";
	print TESTDBLIB <<__E;
#include <$h>
int main(int argc, char *argv[])
{
	DB *dbp = 0;
	dbp = dbopen(0, 0, 0, DB_HASH, 0);
	dbp->close(dbp);
	DBT data;
	dbp->put(dbp, &data, &data, 0);
	dbp->get(dbp, &data, &data, 0);
	return 0;
}
__E
	close TESTDBLIB;
	# attempt to compile
	my $r = system("g++ testdblib.cpp -l$db -o testdblib $compile_line_extra $link_line_extra"); # 2>/dev/null >/dev/null");
	# delete test files
	unlink 'testdblib';
	unlink 'testdblib.cpp';
	return $r == 0; # compilation succeeded
}
# ---- LINUX end --------------------------------------------------------------------------

# check the version of openssl installed
sub test_openssl
{
my $function = $_[0];
open TESTOPENSSL,">testopensslversion.cpp" or die "Can't open openssl test file for writing";
print TESTOPENSSL <<__E;
#include <openssl/evp.h>
int main(int argc, char *argv[])
{
	$function
	return 0;
}
__E
close TESTOPENSSL;
# attempt to compile
my $result = (system("g++ testopensslversion.cpp -lcrypto -o testopensslversion $compile_line_extra $link_line_extra") == 0);
# delete test files
unlink 'testopensslversion';
unlink 'testopensslversion.cpp';
return $result;
}

if(!test_openssl('EVP_CipherInit_ex(0, 0, 0, 0, 0, 0);'))
{
	# failed somehow...

	# is an old version present?
	my $old_version = (test_openssl('EVP_CipherInit(0, 0, 0, 0, 0);'));

	if(!$old_version && !-e '/usr/include/openssl/evp.h')
	{
		# headers not installed?
		print "\n\nERROR: OpenSSL library and headers need to be installed.\n\nSee documentation on web site if you need to add extra search paths.\n\n";
		exit(1);
	}
	else
	{
		if($old_version)
		{
			if($old_version_of_openssl_ok)
			{
				print "\n\nWARNING: Configuring for old version of OpenSSL.\nPerformance will be lower than with version 0.9.7\n\n";
				$compile_line_extra .= '-DPLATFORM_OLD_OPENSSL ';
			}
			else
			{
				# tell user how they might try anyway with an old version of openssl
				print <<__E;

===============================================================================

You appear to have a version of OpenSSL installed which is less than 0.9.7.

It is possible to configure to use this old version, but performance will be
lower than if you have 0.9.7 -- the later version is recommended.

If you wish to use this old version, repeat the configuration with the extra
argument 'allow-old-openssl', like this:

   ./configure allow-old-openssl

* Please do not distribute binary packages with this option enabled.

* Please do not distribute ports which set this option by default.

IMPORTANT: Support for older versions of OpenSSL should be considered
experimental. It is not recommended for production use, but provided to allow
easier evaluation of this software before installing the latest OpenSSL.

===============================================================================

__E
				exit(1);
			}
		}
		else
		{
			print "\n\nERROR: You need to install OpenSSL, preferably at least version 0.9.7.\n\n";
			print "If you believe you have installed OpenSSL, check that the headers are installed\nas well ('dev' packages?)\n\nSee documentation on web site if you need to add extra search paths.\n\n";
			exit(1);
		}
	}
}


# check that the __USE_MALLOC define doesn't break things
my $use_malloc_ok = 1;

if(!$gcc_v3)
{
	open TESTUSEMALLOC,">testusemalloc.cpp" or die "Can't open use_malloc test file for writing";
	print TESTUSEMALLOC <<__E;
#define __USE_MALLOC
#include <string>
int main(int argc, char *argv[])
{
        std::string s;
        s = "test";
}
__E
	close TESTUSEMALLOC;
	$use_malloc_ok = (system("g++ testusemalloc.cpp -o testusemalloc $compile_line_extra $link_line_extra") == 0);
	unlink "testusemalloc.cpp";
	unlink "testusemalloc";
}
if(!$use_malloc_ok)
{
	# don't use __USE_MALLOC then, if it breaks things
	$compile_line_extra .= '-DPLATFORM_STL_USE_MALLOC_BROKEN ';
}


# finished checking the environment
print "done\n\n";


# seed autogen code
print "Seeding autogen code...\n";
open FINDAUTOGEN,"find . -name Makefile.extra |" or die "Can't use find for locating files";
while(<FINDAUTOGEN>)
{
	chomp;
	my $file = $_;
	$file =~ m~\A(.+)/[^/]+\Z~;
	my $dir = $1;
	open FL,$file or die "Can't open $_ for reading";
	my %vars;
	my $do_cmds = 0;
	while(<FL>)
	{
		chomp;
		if(m/\A(.+)\s+=\s+(.+)\Z/)
		{
			# is a variable
			$vars{$1} = $2;
			next;
		}
		next unless m/\S/;
		if(m/AUTOGEN SEEDING/)
		{
			$do_cmds = 1;
		}
		elsif(m/\A\S/)
		{
			$do_cmds = 0 if $do_cmds == 2;
		}
		else
		{
			# command, run it?
			if($do_cmds)
			{
				$do_cmds = 2; # flag something has been done
				
				# subsitute variables, repeatedly
				my $c = $_;
				$c =~ s/\A\s+//;
				while(1)
				{
					my $did_subst = 0;
					
					for my $k (keys %vars)
					{
						$did_subst = 1 if $c =~ s/\$\($k\)/$vars{$k}/g;
					}
					
					last unless $did_subst;
				}
				
				# run command
				die "Couldn't run command $c" unless (0 == system("(cd $dir; $c)"))
			}
		}
	}
	close FL;
}
close FINDAUTOGEN;
print "done\n\n";


# open test mail program template file
my $test_template_file = 'buildenv-testmain-template.h';
open FL,$test_template_file or die "Can't open test template file\n";
my $test_template;
read FL,$test_template,-s $test_template_file;
close FL;


# extra platform defines
my $extra_platform_defines = '';
$extra_platform_defines .= ' -DPLATFORM_GCC3' if $gcc_v3;

# read in module definitions file and prepare directories
open MODULES,"modules.txt" or die "Can't open modules file\n";

# prepare directories...
mkdir "release",0755;
mkdir "debug",0755;

print "Scanning code...\n";

while(<MODULES>)
{
	# clean up line
	chomp; s/\A\s+//; s/#.+\Z//; s/\s+\Z//; s/\s+/ /g;
	next unless m/\S/;
	
	# split up...
	my ($mod, @deps_i) = split / /;
	
	# deps for this platform
	my @deps;
	for(@deps_i)
	{
		my ($dep,$exclude_from) = split /!/;
		$dep =~ s/PSEUDO_DB/$pseudo_db_lib/;
		$dep =~ s/PSEUDO_READLINE/$pseudo_readline_lib/;
		next if $dep =~ m/-lNONE/;
		if($exclude_from =~ m/\A\+(.+)\Z/)
		{
			$exclude_from = $1;
			my $inc = 0;
			for(split /,/,$exclude_from)
			{
				$inc = 1 if $_ eq $build_os
			}
			push @deps,$dep if $inc
		}
		else
		{
			my $inc = 1;
			for(split /,/,$exclude_from)
			{
				$inc = 0 if $_ eq $build_os
			}
			push @deps,$dep if $inc
		}
	}
	
	# check directory exists
	die "Module $mod can't be found\n" unless -d $mod;
	
	# and put in lists
	push @modules,$mod;
	my @md;	# module dependencies
	my @lo; # link line options
	for(@deps)
	{
		if(/\A-l/)
		{
			push @lo,$_
		}
		else
		{
			push @md,$_
		}
	}
	$module_dependency{$mod} = [$implicit_dep,@md];
	$module_library_link_opts{$mod} = [@lo];
	
	# make directories
	my ($s,$d) = split /\//,$mod;
	mkdir "release/$s",0755;
	mkdir "release/$s/$d",0755;
	mkdir "debug/$s",0755;
	mkdir "debug/$s/$d",0755;
}
close MODULES;

# make dirs for implicit dep
mkdir "release/$implicit_dep",0755;
mkdir "debug/$implicit_dep",0755;


# now make a list of all the .h files we can find, recording which module they're in
my %hfiles;
for my $mod (@modules, $implicit_dep)
{
	opendir DIR,$mod;
	my @items = readdir DIR;
	closedir DIR;

	# add in items from autogen directories, and create output directories
	{
		my @autogen_items;
		
		for my $di (@items)
		{
			if($di =~ m/\Aautogen/ && -d "$mod/$di")
			{
				# Read items
				my $d = "$mod/$di";
				opendir DIR,$d;
				my @i = readdir DIR;
				closedir DIR;
				for(@i)
				{
					next if m/\A\./;
					push @autogen_items,"$di/$_"
				}
			}
		}
		@items = (@items, @autogen_items);
	}

	for(grep /\.h\Z/i, @items)
	{
		next if /\A\._/;	# Temp Mac OS Resource hack
		die "Header file $_ already used in module ".$hfiles{$_}."\n" if exists $hfiles{$_};
		$hfiles{$_} = $mod
	}
}

for my $mod (@modules, $implicit_dep)
{
	opendir DIR,$mod;
	for my $h (grep /\.h\Z/i, readdir DIR)
	{	
		next if /\A\._/;	# Temp Mac OS Resource hack

		open FL,"$mod/$h" or die "can't open $mod/$h";
		my $f;
		read FL,$f,-s "$mod/$h";
		close FL;

		while($f =~ m/\#include\s+"([^"]+?)"/g)
		{
			my $i = $1;
			# ignore autogen exceptions
			next if $i =~ m/\Aautogen_.+?Exception.h\Z/;
			# record dependency
			${$header_dependency{$h}}{$i} = 1 if exists $hfiles{$i};
		}
	}
	closedir DIR;
}

print "done\n\nGenerating Makefiles...\n";


# Then write a makefile for each module
for my $mod (@modules, $implicit_dep)
{
	print $mod,"\n";
	
	my ($type,$name) = split /\//,$mod;
	
	# add additional files for tests
	if($type eq 'test')
	{
		my $testmain = $test_template;
		$testmain =~ s/TEST_NAME/$name/g;
		open TESTMAIN,">$mod/_main.cpp" or die "Can't open test main file for $mod for writing\n";
		print TESTMAIN $testmain;
		close TESTMAIN;
		
		# test file...
		sub writetestfile
		{
			my ($filename,$runcmd,$module) = @_;		
			open TESTFILE,">$filename" or die "Can't open test script file for $module for writing\n";
			print TESTFILE "#!/bin/sh\necho TEST: $module\n";
			if(-d "$module/testfiles")
			{
				print TESTFILE <<__E;
echo Removing old test files...
rm -rf testfiles
echo Copying new test files...
cp -p -R ../../../$module/testfiles .
__E
			}
			if(-e "$module/testextra")
			{
				open FL,"$module/testextra" or die "Can't open $module/testextra";
				while(<FL>) {print TESTFILE}
				close FL;
			}
			print TESTFILE "$runcmd\n";
			close TESTFILE;
		}
		
		writetestfile("$mod/_t", './test $1 $2 $3 $4 $5', $mod);
		writetestfile("$mod/_t-gdb", 'gdb ./test', $mod);
		
	}
	
	my @all_deps_for_module;
	{
		# work out what dependencies need to be run
		my @deps_raw;
		sub add_mod_deps
		{
			my ($arr_r,$nm) = @_;
			if($#{$module_dependency{$nm}} >= 0)
			{
				push @$arr_r,@{$module_dependency{$nm}};
				for(@{$module_dependency{$nm}})
				{
					add_mod_deps($arr_r,$_)
				}
			}
		}
		add_mod_deps(\@deps_raw, $mod);
		# and then dedup and reorder them
		my %d_done;
		for(my $a = $#deps_raw; $a >= 0; $a--)
		{
			if(!exists $d_done{$deps_raw[$a]})
			{
				# insert
				push @all_deps_for_module, $deps_raw[$a];
				# mark as done
				$d_done{$deps_raw[$a]} = 1;
			}
		}
	}	
	

	# make include path
	my $include_paths = join(' ',map {'-I../../'.$_} @all_deps_for_module);

	# make target name
	my $end_target = $name;
	$end_target .= '.a' if $type eq 'lib';
	$end_target = 'test' if $type eq 'test';
	# adjust for outdir
	$end_target = '$(OUTDIR)/' . $end_target;

	# start the makefile
	my $mk_name_extra = ($bsd_make)?'':'X';
	open MAKE,">$mod/Makefile".$mk_name_extra or die "Can't open Makefile for $mod\n";
	my $debug_link_extra = ($type ne 'lib')?'../../debug/lib/debug/debug.a':'';
	print MAKE <<__E;
#
# AUTOMATICALLY GENERATED FILE
#    do not edit!
#
#
CXX = g++
AR = ar
RANLIB = ranlib
.ifdef RELEASE
CXXFLAGS = -DNDEBUG -O2 -Wall $include_paths -D$platform_define$extra_platform_defines -DBOX_VERSION="\\"$product_version\\""
OUTBASE = ../../release
OUTDIR = ../../release/$mod
DEPENDMAKEFLAGS = -D RELEASE
VARIENT = RELEASE
.else
CXXFLAGS = -g -Wall $include_paths -D$platform_define$extra_platform_defines -DBOX_VERSION="\\"$product_version\\""
OUTBASE = ../../debug
OUTDIR = ../../debug/$mod
DEPENDMAKEFLAGS =
VARIENT = DEBUG
.endif

__E
	
	# read directory
	opendir DIR,$mod;
	my @items = readdir DIR;
	closedir DIR;
	
	# add in items from autogen directories, and create output directories
	{
		my @autogen_items;
		for my $di (@items)
		{
			if($di =~ m/\Aautogen/ && -d "$mod/$di")
			{
				# Read items
				my $d = "$mod/$di";
				opendir DIR,$d;
				my @i = readdir DIR;
				closedir DIR;
				for(@i)
				{
					next if m/\A\./;
					push @autogen_items,"$di/$_"
				}
				
				# output directories
				mkdir "release/$mod/$di",0755;
				mkdir "debug/$mod/$di",0755;
			}
		}
		@items = (@items, @autogen_items);
	}
	
	# first, obtain a list of depenencies within the .h files
	my %headers;
	for my $h (grep /\.h\Z/i, @items)
	{
		open FL,"$mod/$h";
		my $f;
		read FL,$f,-s "$mod/$h";
		close FL;

		while($f =~ m/\#include\s+"([^"]+?)"/g)
		{
			${$headers{$h}}{$1} = 1 if exists $hfiles{$1};
		}
	}
	
	# ready for the rest of the details...
	my $make;
	
	# then... do the cpp files...
	my @obj_base;
	for my $cpp (@items)
	{
		next unless $cpp =~ m/\A(.+)\.cpp\Z/i;
		next if $cpp =~ /\A\._/;	# Temp Mac OS Resource hack

		# store for later
		my $base = $1;
		push @obj_base,$base;
	
		# get the file...
		open FL,"$mod/$cpp";
		my $f;
		read FL,$f,-s "$mod/$cpp";
		close FL;
		
		my %dep;

		while($f =~ m/\#include\s+"([^"]+?)"/g)
		{
			insert_dep($1, \%dep) if exists $hfiles{$1};
		}
		
		# output filename
		my $out_name = '$(OUTDIR)/'.$base.'.o';
		
		# write the line for this cpp file
		$make .= $out_name.': '.join(' ',$cpp,map
			{ ($hfiles{$_} eq $mod)?$_:'../../'.$hfiles{$_}."/$_" } keys %dep)."\n";
		$make .= "\t\$(CXX) \$(CXXFLAGS) $compile_line_extra -c $cpp -o $out_name\n\n";

	}

	my $has_deps = ($#{$module_dependency{$mod}} >= 0);
# ----- # always has dependencies with debug library
	$has_deps = 1;

	# Depenency stuff
	my $deps_makeinfo;
	if($has_deps)
	{
		if($bsd_make)
		{
			$deps_makeinfo = <<'__E';
.BEGIN::
.ifndef NODEPS
.	if $(.TARGETS) == ""
__E
		}
		else
		{
			# gnu make
			$deps_makeinfo = <<'__E';
.PHONY:	dep_modules
dep_modules:
ifndef NODEPS
ifeq ($(strip $(.TARGETS)),)
__E
		}
		
		# run make for things we require
		for my $dep (@all_deps_for_module)
		{
			$deps_makeinfo .= "\t\t(cd ../../$dep; $make_command$sub_make_options \$(DEPENDMAKEFLAGS) -D NODEPS)\n";
		}
		$deps_makeinfo .= ".\tendif\n.endif\n\n";
	}
	print MAKE $deps_makeinfo if $bsd_make;
	
	# get the list of library things to add -- in order of dependency so things link properly
	my $lib_files = join(' ',map {($_ =~ m/lib\/(.+)\Z/)?('$(OUTBASE)/'.$_.'/'.$1.'.a'):undef} (reverse(@all_deps_for_module)));

	# need to see if the extra makefile fragments require extra object files
	# or include any more makefiles
	my @objs = @obj_base;
	my @makefile_includes;
	
	additional_objects_from_make_fragment("$mod/Makefile.extra", \@objs, \@makefile_includes);
	additional_objects_from_make_fragment("$mod/Makefile.extra.$build_os", \@objs, \@makefile_includes);

	my $o_file_list = join(' ',map {'$(OUTDIR)/'.$_.'.o'} @objs);
	print MAKE $end_target,': ',$o_file_list;
	print MAKE ' dep_modules' if !$bsd_make;
	print MAKE " ",$lib_files unless $type eq 'lib';
	print MAKE "\n";
	
	# stuff to make the final target...
	if($type eq 'lib')
	{
		# make a library archive...
		print MAKE "\t(echo -n > $end_target; rm $end_target)\n";
		print MAKE "\t\$(AR) -q $end_target $o_file_list\n";
		print MAKE "\t\$(RANLIB) $end_target\n";
	}
	else
	{
		# work out library options
		# need to be... least used first, in absolute order they appear in the modules.txt file
		my @libops;
		sub libops_fill
		{
			my ($m,$r) = @_;
			push @$r,$_ for(@{$module_library_link_opts{$m}});
			libops_fill($_,$r) for(@{$module_dependency{$m}});
		}
		libops_fill($mod,\@libops);
		my $lo = '';
		my %ldone;
		for(@libops)
		{
			next if exists $ldone{$_};
			$lo .= ' '.$_;
			$ldone{$_} = 1;
		}
	
		# link line...
		print MAKE "\t\$(CXX) $link_line_extra -o $end_target $o_file_list $lib_files$lo\n";
	}
	# tests need to copy the test file over
	if($type eq 'test')
	{
		print MAKE "\tcp _t \$(OUTDIR)/t\n\tchmod u+x \$(OUTDIR)/t\n";
		print MAKE "\tcp _t-gdb \$(OUTDIR)/t-gdb\n\tchmod u+x \$(OUTDIR)/t-gdb\n";
	}
	# dependency line?
	print MAKE "\n";

	# module dependcies for GNU make?
	print MAKE $deps_makeinfo if !$bsd_make;
	
	# print the rest of the file
	print MAKE $make,"\n";
	
	# and a clean target
	print MAKE "clean:\n\t-rm -rf \$(OUTDIR)/*\n.\tifndef SUBCLEAN\n";
	for my $dep (@all_deps_for_module)
	{
		print MAKE "\t(cd ../../$dep; $make_command \$(DEPENDMAKEFLAGS) -D SUBCLEAN clean)\n";
	}
	print MAKE ".\tendif\n";
	
	# include any extra stuff
	print MAKE "\n\n";
	if(-e "$mod/Makefile.extra")
	{
		print MAKE ".include <Makefile.extra>\n\n";
	}
	if(-e "$mod/Makefile.extra.$build_os")
	{
		print MAKE ".include <Makefile.extra.$build_os>\n\n";
	}
	for(@makefile_includes)
	{
		print MAKE ".include <$_>\n\n";
	}
	
	# and finally a target for rebuilding the build system
	print MAKE "\nbuildsystem:\n\t(cd ../..; perl ./makebuildenv.pl)\n\n";
	
	close MAKE;

	if(!$bsd_make)
	{
		# need to post process this into a GNU makefile
		open MAKE,">$mod/Makefile";
		open MAKEB,"$mod/MakefileX";

		while(<MAKEB>)
		{
			s/\A\.\s*(ifdef|else|endif|ifndef)/$1/;
			s/\A\.\s*include\s+<(.+?)>/include $1/;
			s/-D\s+(\w+)/$1=1/;
			print MAKE;
		}

		close MAKEB;
		close MAKE;
		unlink "$mod/MakefileX";
	}
}

print "\nType 'cd <module_dir>; $make_command' to build a module\n\n";

if(!$use_malloc_ok)
{
	print <<__E;

================
WARNING: The implementation of the C++ STL on this platform may have a flaw
which causes it to apparently leak memory, and this flaw cannot be worked
around.

When running the daemons, check their memory usage does not constantly
increase. The STL flaw can cause excessive memory use.
================

__E
}

sub insert_dep
{
	my ($h,$dep_r) = @_;
	
	# stop random recusion
	return if exists $$dep_r{$h};
	
	# insert more depencies
	insert_dep($_,$dep_r) for keys %{$header_dependency{$h}};

	# mark this one as a dependency
	$$dep_r{$h} = 1;
}


sub additional_objects_from_make_fragment
{
	my ($fn,$objs_r,$include_r) = @_;
	
	if(-e $fn)
	{
		open FL,$fn or die "Can't open $fn";
		
		while(<FL>)
		{
			chomp;
			if(m/link-extra:\s*(.+)\Z/)
			{
				my @o = split /\s+/,$1;
				for(@o)
				{
					push @$objs_r,$1 if m/\A(.+)\.o\Z/;
				}
			}
			elsif(m/include-makefile:\s*(\S+)/)
			{
				push @$include_r,$1
			}
		}
		
		close FL;
	}
}

